#include <asm/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <net/if.h>
#include <netinet/in.h>

#include <linux/netlink.h>
#include <linux/rtnetlink.h>

int open_netlink()
{
  int sock = socket(PF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
  struct sockaddr_nl addr;

  memset((void *)&addr, 0, sizeof(addr));

  if (sock<0)
    return sock;
  addr.nl_family = AF_NETLINK;
  addr.nl_pid = getpid();
  addr.nl_groups = RTMGRP_LINK;

  if (bind(sock,(struct sockaddr *)&addr,sizeof(addr))<0)
    return -1;
  return sock;
}

static int netlink_link_state(struct sockaddr_nl *nl, struct nlmsghdr *msg)
{
  struct ifinfomsg *infomsg = NLMSG_DATA(msg);
  struct rtattr *rta;
  int len = msg->nlmsg_len;

  rta = IFLA_RTA(infomsg);

  while (RTA_OK(rta, len)){
    if (rta->rta_type == IFLA_MTU) {
      unsigned int *mtu = NULL;
      mtu = RTA_DATA(rta);
      printf("       mtu = %u\n", *mtu);
    }
    else if( rta->rta_type == IFLA_IFNAME ) {
      printf("       ifname = %s\n", (char *)RTA_DATA(rta));
    }
    rta = RTA_NEXT(rta, len);
  }

  return 0;
}

static int msg_handler(struct sockaddr_nl *nl, struct nlmsghdr *msg)
{
  struct ifinfomsg *ifi=NLMSG_DATA(msg); /* for RTM_{NEW,DEL}ADDR */
  char val[1024];

  printf("new msg. msg_type = %d\n", msg->nlmsg_type);

  if( msg->nlmsg_type == RTM_NEWLINK) {
      if_indextoname(ifi->ifi_index, val);
      printf("msg_handler: RTM_NEWLINK: ifi %s\n", val);
      netlink_link_state(nl, msg);
  }

  return 0;
}

int read_event(int sockint, int (*msg_handler)(struct sockaddr_nl *,
                                               struct nlmsghdr *))
{
  int status;
  int ret = 0;
  char buf[4096];
  struct iovec iov = { buf, sizeof buf };
  struct sockaddr_nl snl;
  struct msghdr msg = { (void*)&snl, sizeof snl, &iov, 1, NULL, 0, 0};
  struct nlmsghdr *h;

  status = recvmsg(sockint, &msg, 0);

  if(status < 0)
    {
      /* Socket non-blocking so bail out once we have read everything */
      if (errno == EWOULDBLOCK || errno == EAGAIN)
        return ret;

      /* Anything else is an error */
      printf("read_netlink: Error recvmsg: %d\n", status);
      perror("read_netlink: Error: ");
      return status;
    }

  if(status == 0)
    {
      printf("read_netlink: EOF\n");
    }

  /* We need to handle more than one message per 'recvmsg' */
  for(h = (struct nlmsghdr *) buf; NLMSG_OK (h, (unsigned int)status);
      h = NLMSG_NEXT (h, status))
    {
      /* Finish reading */
      if (h->nlmsg_type == NLMSG_DONE)
        return ret;

      /* Message is some kind of error */
      if (h->nlmsg_type == NLMSG_ERROR)
        {
          printf("read_netlink: Message is an error - decode TBD\n");
          return -1; // Error
        }

      /* Call message handler */
      if(msg_handler)
        {
          ret = (*msg_handler)(&snl, h);
          if(ret < 0)
            {
              printf("read_netlink: Message handler error %d\n", ret);
              return ret;
            }
        }
      else
        {
          printf("read_netlink: Error NULL message handler\n");
          return -1;
        }
    }

  return ret;
}

int
main(int argc, char *argv[])
{
  int sock_fd;

  printf("create socket\n");

  sock_fd = open_netlink();
  if (sock_fd < 0) {
    printf("socket: %s\n", strerror(errno));
    return 1;
  }

  printf("read event\n");

  while( 1 )
    read_event(sock_fd, msg_handler);


  return 0;
}
